/**
 * xethya-random-mtw
 *
 * Copyright © 2016 Joel A. Villarreal Bertoldi. All rights reserved.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE.txt file in the root directory of this source tree.
 */

import { expect } from 'chai';
import MersenneTwister from '../src/MersenneTwister';

describe('MersenneTwister', () => {

  it('should initialize correctly with either method', () => {
    const mt = new MersenneTwister();

    expect(mt.MT).to.satisfy(arr => arr.every(num => !isNaN(num)));
    expect(mt.MTI).to.not.equal(0);

    mt.initializeByArray([15, 17, 3, 58]);
    expect(mt.MT[0]).to.equal(0x80000000);
  });

  it('should initialize correctly with either method, '
    + 'with an explicitly-set undefined value', () => {
    const mt = new MersenneTwister(undefined);

    expect(mt.MT).to.satisfy(arr => arr.every(num => !isNaN(num)));
    expect(mt.MTI).to.not.equal(0);

    mt.initializeByArray([15, 17, 3, 58]);
    expect(mt.MT[0]).to.equal(0x80000000);
  });

  it('should initialize correctly with a positive seed', () => {
    const mt = new MersenneTwister(Math.abs(Math.floor(Math.random() * 10000)));
    expect(mt.MT).to.satisfy(arr => arr.every(num => !isNaN(num)));
    expect(mt.MTI).to.not.equal(0);
  });

  it('should initialize correctly with a negative seed (it becomes Math.abs()\'ed)', () => {
    const mt = new MersenneTwister(-Math.abs(Math.floor(Math.random() * 10000)));
    expect(mt.MT).to.satisfy(arr => arr.every(num => !isNaN(num)));
    expect(mt.MTI).to.not.equal(0);
  });

  it('should fail to initialize correctly with an empty list', () => {
    const mt = new MersenneTwister();
    expect(() => mt.initializeByArray([])).to.throw(/initKeyArray/);
  });

  it('should allow to reinitialize the generator at any given time without '
    + 'reinstantiating, with a positive seed', () => {
    const mt = new MersenneTwister();
    expect(mt.MT).to.satisfy(arr => arr.every(num => !isNaN(num)));
    expect(mt.MTI).to.not.equal(0);
    const newSeed = Math.abs(Math.floor(Math.random() * 10000));
    mt.initializeRandomGenerator(newSeed);
    expect(mt.MT[0]).to.equal(newSeed >> 0);
  });

  it('should allow to reinitialize the generator at any given time without '
    + 'reinstantiating, with a negative seed', () => {
    const mt = new MersenneTwister();
    expect(mt.MT).to.satisfy(arr => arr.every(num => !isNaN(num)));
    expect(mt.MTI).to.not.equal(0);
    const newSeed = -Math.abs(Math.floor(Math.random() * 10000));
    mt.initializeRandomGenerator(newSeed);
    expect(mt.MT[0]).to.equal(Math.abs(newSeed) >> 0);
  });

  it('should not allow to reinitialize the generator at any given '
    + 'time if no seed is provided', () => {
    const mt = new MersenneTwister();
    expect(mt.MT).to.satisfy(arr => arr.every(num => !isNaN(num)));
    expect(mt.MTI).to.not.equal(0);
    expect(() => mt.initializeRandomGenerator()).to.throw(/A seed number is required/);
  });

  it('should generate a value with return type Number', () => {
    const mt = new MersenneTwister();
    expect(mt.generateRandomInteger()).to.be.a('number');
    expect(mt.generateRandomInteger31()).to.be.a('number');
    expect(mt.generateRandomReal53BitResolution()).to.be.a('number');
    expect(mt.generateRandomReal()).to.be.a('number');
    expect(mt.generateRandomReal3()).to.be.a('number');
    expect(mt.generateRandom()).to.be.a('number');
  });

  it('should generate non-negative numbers', () => {
    const mt = new MersenneTwister();
    expect(mt.generateRandomInteger31()).to.be.at.least(0);
    expect(mt.generateRandomReal()).to.be.at.least(0);
    expect(mt.generateRandom()).to.be.at.least(0);
    expect(mt.generateRandomReal3()).to.be.at.least(0);
    expect(mt.generateRandomReal53BitResolution()).to.be.at.least(0);
  });

  it('should generate random numbers between 0 and 1', () => {
    const mt = new MersenneTwister();
    expect(mt.generateRandomReal()).to.be.within(0, 1);
    expect(mt.generateRandom()).to.be.within(0, 1);
    expect(mt.generateRandomReal3()).to.be.within(0, 1);
  });

});
